/*
	GLSL Shader abstraction
 
	MIT License

	Copyright (c) 2009, Markus Broecker <mbroecker@gmail.com>

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

#include "Shader.h"
//#include "Textures.h"
#include <OpenGL/OpenGL.h>

#include <vector>
#include <fstream>
#include <iostream>
#include <cstring>


Shader::Shader( const std::string& vpFile, const std::string& fpFile)
{
	//std::clog << "Loading shader, v: \"" << vpFile << ", f: \"" << fpFile << "\"" << std::endl;
	load( vpFile, fpFile );
}

Shader::~Shader()
{
	glDeleteProgram( mProgram );
	glDeleteShader( mVS );
	glDeleteShader( mFS );	
}

void Shader::bind() const
{
	glUseProgram( mProgram );
}

void Shader::disable() const
{
	glUseProgram( 0 );
}

int Shader::getUniform(const std::string& name) const
{
	return glGetUniformLocation( mProgram, name.c_str() );
}

void Shader::setUniform(const std::string& name, int i) const
{
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), i );
}

void Shader::setUniform(const std::string& name, float f) const
{
	glUniform1f( glGetUniformLocation( mProgram, name.c_str() ), f );
}

void Shader::setUniform(const std::string& name, float x, float y) const
{	
	glUniform2f( glGetUniformLocation( mProgram, name.c_str() ), x, y);
}

void Shader::setUniform(const std::string& name, float x, float y, float z) const
{
	glUniform3f( glGetUniformLocation( mProgram, name.c_str() ), x, y, z );
}

void Shader::setUniform(const std::string& name, float x, float y, float z, float w) const
{
	glUniform4f( glGetUniformLocation( mProgram, name.c_str() ), x, y, z, w);
}

void Shader::setMatrix4(const std::string& name, const float* matrix, bool transpose) const
{
	glUniformMatrix4fv(glGetUniformLocation(mProgram, name.c_str()), 1, transpose, matrix);
}

void Shader::setMatrix4(const std::string& name, const double* matrix, bool transpose) const
{
	float temp[16];
	for (size_t i = 0; i < 16; ++i)
		temp[i] = static_cast<float>(matrix[i]);
	this->setMatrix4(name, temp, transpose);
}

void Shader::setTexture1D(const std::string& name, unsigned int textureId, unsigned int textureUnit) const
{
	glActiveTexture(GL_TEXTURE0 + textureUnit);
	glBindTexture(GL_TEXTURE_1D, textureId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), textureUnit );
}

void Shader::setTexture2D(const std::string& name, unsigned int textureId, unsigned int textureUnit) const
{
	glActiveTexture(GL_TEXTURE0 + textureUnit);
	glBindTexture(GL_TEXTURE_2D, textureId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), textureUnit );
}

void Shader::setTexture3D(const std::string& name, unsigned int textureId, unsigned int textureUnit) const
{
	glActiveTexture(GL_TEXTURE0 + textureUnit);
	glBindTexture(GL_TEXTURE_3D, textureId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), textureUnit );
}

void Shader::setTextureCube(const std::string& name, unsigned int texId, unsigned int unit) const
{
	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(GL_TEXTURE_CUBE_MAP, texId);
	glUniform1i( glGetUniformLocation( mProgram, name.c_str() ), unit );
}

void Shader::loadShader(const std::string& fileName, GLenum type)
{
	GLint result, length;
	std::vector<std::string>	contents;
	std::ifstream file;
	file.open( fileName.c_str() );
	
	if (!file.is_open())
	{
		std::cerr << "Error opening file " << fileName << "!\n";
		return;
	}
	
	std::string line;
	while (!file.eof())
	{
		std::getline( file, line );
		line.append("\n");
		contents.push_back( line );	
	}
		
	file.close();

	int current = 0;
	if (type == GL_VERTEX_SHADER)
		current = mVS = glCreateShader( GL_VERTEX_SHADER );
	else
		current = mFS = glCreateShader( GL_FRAGMENT_SHADER );
		
	
	const char **glLines =  new const char *[contents.size()];
	for (unsigned int i = 0; i < contents.size(); ++i)
		glLines[i] = contents[i].c_str();
						
	glShaderSource( current, contents.size(), glLines , 0 );
	delete [] glLines;

	glCompileShader( current );
	
	glGetShaderiv(current, GL_COMPILE_STATUS, &result);

	if ( result == GL_FALSE ){
	    glGetShaderiv(current, GL_INFO_LOG_LENGTH, &length);
	    GLchar *buffer = new GLchar[length];
	    glGetShaderInfoLog( current, length, NULL, buffer );
	    std::cerr << (type==GL_VERTEX_SHADER?" Vertex":"Fragment") << "shader \"" << fileName << "\" compile error: " << buffer;
	    std::cerr << std::endl;
	    delete buffer;
	}
}

void Shader::load(const std::string& vpFile, const std::string& fpFile)
{
	GLint result, length;

	loadShader( vpFile, GL_VERTEX_SHADER );
	loadShader( fpFile, GL_FRAGMENT_SHADER );

	mProgram = glCreateProgram();
	glAttachShader( mProgram, mVS );
	glAttachShader( mProgram, mFS );;
	glLinkProgram( mProgram );

	glGetProgramiv(mProgram, GL_LINK_STATUS, &result);
	if( result == GL_FALSE )
	{
	    glGetProgramiv(mProgram, GL_INFO_LOG_LENGTH, &length);
	    GLchar *buffer = new GLchar[length];
	    glGetProgramInfoLog( mProgram, length, NULL, buffer );
	    std::cerr << "Shader link error: " << buffer << std::endl;
	    delete [] buffer;
	}

	glValidateProgram( mProgram );
	glGetProgramiv(mProgram, GL_VALIDATE_STATUS, &result);
	if (result == GL_FALSE)
	{
		std::cerr << "Error validating shader\n";
	}
}
