/*
 Volume Rendering Demo
 
 MIT License
 
 Copyright (c) 2011, Markus Broecker <mbroecker@gmail.com>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

#include <iostream>
#include <algorithm>

#include <OpenGL/OpenGL.h>
#include <GLUT/GLUT.h>
#include <cmath>

#include "VolumetricData.h"
#include "Shader.h";
#include "OrbitCamera.h"

static SlicedVolume volume;
static Shader* volumeShader = 0;

static OrbitCamera	camera;

static float threshold = 0.2f;
static float low_density[3] = { 0.3, 0.8, 0.3 };
static float high_density[3] = { 0.8, 0.2, 0.2 };
static float transparency = 0.f;

static float animation = 0.f;
static float inline radians(float deg) { return deg*0.01745;}

static const unsigned int WIDTH = 800;
static const unsigned int HEIGHT = 800;

static void display()
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45, 1, 0.1, 500);
	
	
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	
	float x = sin(radians(animation)) * 10 + 2.5;
	float z = cos(radians(animation)) * 10 + 2.5;
	gluLookAt(x, 10, z, 2.5, 2, 2.5, 0, 1, 0);
	
	
	camera.setup();
	
	
	// draw a ground grid
	glColor3f(0.4, 0.4, 0.4);
	glBegin(GL_LINES);
	for (int i = -20; i <= 20; i+=5)
	{
		glVertex3i(-20, 0, i);
		glVertex3i( 20, 0, i);
		glVertex3i(i, 0, -20);
		glVertex3i(i, 0,  20);
	}
	glEnd();
	
	glScalef(5, 5, 5);
	volumeShader->bind();
	volumeShader->setUniform("slice", 0);
	volumeShader->setUniform("threshold", threshold);
	volumeShader->setUniform("low_density", low_density[0], low_density[1], low_density[2]);
	volumeShader->setUniform("high_density", high_density[0], high_density[1], high_density[2]);
	volumeShader->setUniform("transparency", transparency);
	volume.draw();
	
	volumeShader->disable();
	
	
	glutSwapBuffers();
}

static void idle()
{
	static unsigned long oldtime = 0;
	unsigned long time = glutGet(GLUT_ELAPSED_TIME);
	float dt = float(time-oldtime) / 1000.f;
	oldtime = time;
	
	animation += dt*10;
	
	
	glutPostRedisplay();
}

static void keyboard(unsigned char key, int x, int y)
{
	if (key == 27)
		exit(0);
	
	if (key == 'a')
		threshold = std::min(1.f, threshold+0.1f);
	if (key == 'z')
		threshold = std::max(0.f, threshold-0.1f);
	if (key == 't')
		transparency = 1.f-transparency;
	
	if (key == 's')
	{
		camera.radius = std::max(1.f, camera.radius*0.7f);
		camera.updatePosition();
	}
	
	if (key == 'x')
	{
		camera.radius = std::min(15.f, camera.radius*1.4f);
		camera.updatePosition();
	}
		
}

static void motion(int x, int y)
{
	static int oldx = 0;
	static int oldy = 0;
	
	
	float dx = float(x-oldx) / WIDTH;
	float dy = float(y-oldy) / HEIGHT;
	
	
	//std::clog << "dx:" << dx << " dy:" << dy << std::endl;

	camera.rotateTo(dx, dy);
	

}

int main (int argc, char * argv[]) 
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA|GLUT_DEPTH|GLUT_DOUBLE);
	glutInitWindowSize(WIDTH, HEIGHT);
	
	glutCreateWindow("Volume Renderer");
	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutKeyboardFunc(keyboard);
	glutMotionFunc(motion);
	
	glEnable(GL_DEPTH_TEST);
	glClearColor(0, 0, 0.2, 0);

	if (argc != 5)
	{
		std::cerr << "Usage: " << argv[0] << " <filename> <width> <height> <depth>\n";
		return 1;
	}
		
	volume.loadRAW(argv[1], atoi(argv[2]), atoi(argv[3]), atoi(argv[4]));

	//volume.loadRAW("BostonTeapot.raw", 256, 256, 178);
	//volume.loadRAW("aneurism.raw", 256, 256, 256);
	//volume.loadRAW("lobster.raw", 301, 324, 256);
	//volume.loadRAW("bonsai.raw", 256, 256, 256);
	//volume.loadRAW("skull.raw", 256, 256, 256);
	
	volumeShader = new Shader("volumetric.vert", "volumetric.frag");
	
	camera.ratio = 1.f;
	camera.near = 0.1;
	camera.far = 100.f;
	camera.radius = 10.f;
	camera.target.set(2.5, 2.5, 2.5);
	
	glutMainLoop();
	
	return 0;
}
